from django.db import models
from django.utils.translation import gettext_lazy as _


class IXP(models.Model):
    name = models.CharField(max_length=150)
    long_name = models.CharField(max_length=200)
    city = models.CharField(max_length=200)
    website = models.URLField(null=True)
    active_status = models.BooleanField(default=True)
    peeringdb_id = models.IntegerField(null=True)
    country_code = models.CharField(max_length=2)
    created = models.DateTimeField()
    last_updated = models.DateTimeField()
    last_active = models.DateTimeField(null=True)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("Internet Exchange Point")
        verbose_name_plural = _("Internet Exchange Points")

        constraints = [
            models.UniqueConstraint(fields=['peeringdb_id'], name='ixp_tracker_unique_ixp_peeringdb_id')
        ]


class ASN(models.Model):
    NETWORK_TYPE_CHOICES = [
        ('access', 'Access'),
        ('transit', 'Transit'),
        ('content', 'Content'),
        ('enterprise', 'Enterprise'),
        ('NREN', 'Research and Education'),
        ('non-profit', 'Non-profit'),
        ('not-disclosed', 'Not Disclosed'),
        ('other', 'Other'),
    ]
    name = models.CharField(max_length=500)
    number = models.IntegerField()
    peeringdb_id = models.IntegerField(null=True)
    network_type = models.CharField(max_length=200, choices=NETWORK_TYPE_CHOICES, default='not-disclosed')
    registration_country_code = models.CharField(max_length=2)
    created = models.DateTimeField()
    last_updated = models.DateTimeField()

    def __str__(self):
        return "AS" + str(self.number)

    class Meta:
        verbose_name = "AS Number"
        verbose_name_plural = "AS Numbers"

        constraints = [
            models.UniqueConstraint(fields=['number'], name='ixp_tracker_unique_as_number')
        ]


class IXPMember(models.Model):
    ixp = models.ForeignKey(IXP, on_delete=models.CASCADE)
    asn = models.ForeignKey(ASN, on_delete=models.CASCADE)
    last_updated = models.DateTimeField()
    last_active = models.DateTimeField(null=True)

    def __str__(self):
        return f"{self.ixp.name} - {self.asn.name}"

    class Meta:
        verbose_name = "IXP Member"
        verbose_name_plural = "IXP Members"

        constraints = [
            models.UniqueConstraint(fields=['ixp', 'asn'], name='ixp_tracker_unique_ixp_membership')
        ]


class IXPMembershipRecord(models.Model):
    member = models.ForeignKey(IXPMember, on_delete=models.CASCADE, related_name="memberships")
    start_date = models.DateField()
    is_rs_peer = models.BooleanField(default=False)
    speed = models.IntegerField(null=True)
    end_date = models.DateField(null=True)

    def __str__(self):
        return f"Membership record {self.member} from {self.start_date} to {self.end_date}"

    class Meta:
        verbose_name = "IXP Membership record"
        verbose_name_plural = "IXP Membership records"


class StatsPerIXP(models.Model):
    ixp = models.ForeignKey(IXP, on_delete=models.CASCADE)
    stats_date = models.DateField()
    capacity = models.FloatField()
    members = models.IntegerField()
    local_asns_members_rate = models.FloatField()

    def __str__(self):
        return f"{self.ixp.name} - {self.stats_date}"

    class Meta:
        verbose_name = "IXP stats"

        constraints = [
            models.UniqueConstraint(fields=['ixp', 'stats_date'], name='ixp_tracker_unique_ixp_stats')
        ]

class StatsPerCountry(models.Model):
    country_code = models.CharField(max_length=2)
    stats_date = models.DateField()
    asn_count = models.IntegerField()
    member_count = models.IntegerField()
    asns_ixp_member_rate = models.FloatField()
    total_capacity = models.FloatField()

    def __str__(self):
        return f"{self.country_code}-{self.stats_date}-{self.asn_count}-{self.member_count}"

    class Meta:
        verbose_name = "Per-country stats"

        constraints = [
            models.UniqueConstraint(fields=['country_code', 'stats_date'], name='ixp_tracker_unique_per_country_stats')
        ]



import json
from datetime import datetime
import re

import pytest
import responses

from ixp_tracker.conf import DATA_ARCHIVE_URL
from ixp_tracker.importers import import_data
from ixp_tracker.management.commands.ixp_tracker_import import DefaultASNGeoLookup
from ixp_tracker.models import ASN, IXP, IXPMember

pytestmark = pytest.mark.django_db


def test_with_no_data_returned_does_nothing():
    backfill_date = datetime(year=2024, month=1, day=1)
    with responses.RequestsMock() as rsps:
        rsps.get(
            url=DATA_ARCHIVE_URL.format(year=backfill_date.year, month=backfill_date.month, day=backfill_date.day),
            body=json.dumps({"ix": {"data": []}, "net": {"data": []}, "netixlan": {"data": []}}),
        )
        import_data(DefaultASNGeoLookup(), False, backfill_date)

    ixps = IXP.objects.all()
    assert len(ixps) == 0

    asns = ASN.objects.all()
    assert len(asns) == 0

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_handles_malformed_archives():
    backfill_date = datetime(year=2024, month=1, day=1)
    with responses.RequestsMock() as rsps:
        rsps.get(
            url=DATA_ARCHIVE_URL.format(year=backfill_date.year, month=backfill_date.month, day=backfill_date.day),
            body=json.dumps({}),
        )
        import_data(DefaultASNGeoLookup(), False, backfill_date)

    ixps = IXP.objects.all()
    assert len(ixps) == 0

    asns = ASN.objects.all()
    assert len(asns) == 0

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_handles_single_quoted_json():
    backfill_date = datetime(year=2024, month=1, day=1)
    with responses.RequestsMock() as rsps:
        rsps.get(
            url=DATA_ARCHIVE_URL.format(year=backfill_date.year, month=backfill_date.month, day=backfill_date.day),
            body="{'ix': {'data': []}, 'net': {'data': []}, 'netixlan': {'data': []}}",
        )
        import_data(DefaultASNGeoLookup(), False, backfill_date)

    ixps = IXP.objects.all()
    assert len(ixps) == 0

    asns = ASN.objects.all()
    assert len(asns) == 0

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_queries_for_every_day_of_month():
    backfill_date = datetime(year=2024, month=1, day=1)
    with responses.RequestsMock() as rsps:
        data_url = DATA_ARCHIVE_URL.format(year=backfill_date.year, month=backfill_date.month, day=backfill_date.day)
        data_url = data_url.replace("01.json", r'[0-9]{2}\.json')
        rsps.get(
            url=re.compile(data_url),
            status=404,
        )
        import_data(DefaultASNGeoLookup(), False, backfill_date)

    ixps = IXP.objects.all()
    assert len(ixps) == 0

    asns = ASN.objects.all()
    assert len(asns) == 0

    members = IXPMember.objects.all()
    assert len(members) == 0


def test_adds_all_data():
    backfill_date = datetime(year=2024, month=1, day=1)
    with responses.RequestsMock() as rsps:
        rsps.get(
            url=DATA_ARCHIVE_URL.format(year=backfill_date.year, month=backfill_date.month, day=backfill_date.day),
            body=json.dumps(
                {
                    "ix":
                        {
                            "data":
                                [
                                    {
                                        "id": 1,
                                        "name": "City IX",
                                        "name_long": "City Internet Exchange Point",
                                        "city": "City",
                                        "country": "AF",
                                        "website": "http://example.com",
                                        "created": "2019-08-24T14:15:22Z",
                                        "updated": "2019-08-24T14:15:22Z",
                                    }
                                ]
                        },
                    "net":
                        {
                            "data":
                                [
                                    {
                                        "id": 3,
                                        "asn": 6543,
                                        "name": "New ASN",
                                        "info_type": "non-profit",
                                        "created": "2019-08-24T14:15:22Z",
                                        "updated": "2019-08-24T14:15:22Z",
                                    }

                                ]
                        },
                    "netixlan":
                        {
                            "data":
                                [
                                    {
                                        "asn": 6543,
                                        "ix_id": 1,
                                        "created": "2019-08-24T14:15:22Z",
                                        "updated": "2019-08-24T14:15:22Z",
                                        "is_rs_peer": True,
                                        "speed": 10000,
                                    }
                                ]
                        }
                }
            ),
        )
        import_data(DefaultASNGeoLookup(), False, backfill_date)

    ixps = IXP.objects.all()
    assert len(ixps) == 1

    asns = ASN.objects.all()
    assert len(asns) == 1

    members = IXPMember.objects.all()
    assert len(members) == 1



import importlib
import logging
import traceback
from datetime import datetime, timezone
from typing import List

from django.core.management import BaseCommand

from ixp_tracker.conf import IXP_TRACKER_GEO_LOOKUP_FACTORY
from ixp_tracker.importers import ASNGeoLookup, import_data
from ixp_tracker.stats import generate_stats

logger = logging.getLogger("ixp_tracker")


class DefaultASNGeoLookup(ASNGeoLookup):

    def get_iso2_country(self, asn: int, as_at: datetime) -> str:
        return "ZZ"

    def get_status(self, asn: int, as_at: datetime) -> str:
        return "assigned"

    def get_asns_for_country(self, country: str, as_at: datetime) -> List[int]:
        pass


def load_geo_lookup(geo_lookup_name):
    if geo_lookup_name is not None:
        lookup_parts = geo_lookup_name.split(".")
        factory_name = lookup_parts.pop()
        module_name = ".".join(lookup_parts)
        logger.debug("Trying to load geo lookup", extra={"module_name": module_name, "factory": factory_name})
        if module_name and factory_name:
            imported_module = importlib.import_module(module_name)
            factory = getattr(imported_module, factory_name)
            return factory()
    return None


class Command(BaseCommand):
    help = "Updates IXP data"

    def add_arguments(self, parser):
        parser.add_argument("--reset-asns", action="store_true", default=False, help="Do a full reset of ASNs rather than incremental update")
        parser.add_argument("--backfill", type=str, default=None, help="The month you would like to backfill data for")

    def handle(self, *args, **options):
        try:
            logger.debug("Importing IXP data")
            geo_lookup = load_geo_lookup(IXP_TRACKER_GEO_LOOKUP_FACTORY) or DefaultASNGeoLookup()
            reset = options["reset_asns"]
            backfill_date = options["backfill"]
            processing_date = None
            if backfill_date is None:
                import_data(geo_lookup, reset)
            else:
                processing_date = datetime.strptime(backfill_date, "%Y%m").replace(tzinfo=timezone.utc)
                if reset:
                    logger.warning("The --reset option has no effect when running a backfill")
                import_data(geo_lookup, False, processing_date)

            logger.debug("Generating stats")
            generate_stats(geo_lookup, processing_date)
            logger.info("Import finished")
        except Exception as e:
            logging.error("Failed to import data", extra={"error": str(e), "trace": traceback.format_exc()})



from datetime import datetime
from typing import List

import pytest

from ixp_tracker.models import IXPMember, IXPMembershipRecord, StatsPerIXP
from ixp_tracker.stats import calculate_local_asns_members_rate, generate_stats
from tests.test_members_import import create_asn_fixture, create_ixp_fixture

pytestmark = pytest.mark.django_db

class TestLookup:

    def __init__(self, default_status: str = "assigned"):
        self.default_status = default_status

    def get_iso2_country(self, asn: int, as_at: datetime) -> str:
        pass

    def get_status(self, asn: int, as_at: datetime) -> str:
        pass

    def get_asns_for_country(self, country: str, as_at: datetime) -> List[int]:
        return [12345, 446, 789, 5050, 54321]


def test_with_no_data_generates_no_stats():
    generate_stats(TestLookup())

    stats = StatsPerIXP.objects.all()
    assert len(stats) == 0


def test_generates_capacity_and_member_count():
    ixp = create_ixp_fixture(123)
    create_member_fixture(ixp, 12345, 500)
    create_member_fixture(ixp, 67890, 10000)

    generate_stats(TestLookup())

    stats = StatsPerIXP.objects.all()
    assert len(stats) == 1
    ixp_stats = stats.first()
    assert ixp_stats.members == 2
    assert ixp_stats.capacity == 10.5


def test_generates_stats_for_first_of_month():
    create_ixp_fixture(123)

    generate_stats(TestLookup(), datetime(year=2024, month=2, day=10))

    stats = StatsPerIXP.objects.all()
    assert len(stats) == 1
    ixp_stats = stats.first()
    assert ixp_stats.stats_date == datetime(year=2024, month=2, day=1).date()


def test_does_not_count_members_marked_as_left():
    ixp = create_ixp_fixture(123)
    create_member_fixture(ixp, 12345, 500)
    create_member_fixture(ixp, 67890, 10000, datetime(year=2024, month=4, day=1).date())

    generate_stats(TestLookup())

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.members == 1
    assert ixp_stats.capacity == 0.5


def test_does_not_count_members_not_yet_created():
    ixp = create_ixp_fixture(123)
    create_member_fixture(ixp, 12345, 500, member_since=datetime(year=2024, month=1, day=1).date())
    create_member_fixture(ixp, 67890, 10000, member_since=datetime(year=2024, month=4, day=1).date())

    generate_stats(TestLookup(), datetime(year=2024, month=2, day=1))

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.members == 1
    assert ixp_stats.capacity == 0.5


def test_does_not_count_ixps_not_yet_created():
    ixp = create_ixp_fixture(123)
    ixp.created = datetime(year=2024, month=4, day=1)
    ixp.save()
    create_member_fixture(ixp, 12345, 500)
    create_member_fixture(ixp, 67890, 10000)

    generate_stats(TestLookup(), datetime(year=2024, month=2, day=1))

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats is None


def test_saves_local_asns_members_rate():
    ixp_one = create_ixp_fixture(123, "CH")
    create_member_fixture(ixp_one, 12345, 500, asn_country="CH")
    create_member_fixture(ixp_one, 67890, 10000, asn_country="CH")
    ixp_two = create_ixp_fixture(456, "CH")
    create_member_fixture(ixp_two, 54321, 500, asn_country="CH")
    create_member_fixture(ixp_two, 9876, 10000, asn_country="CH")

    generate_stats(TestLookup())

    ixp_stats = StatsPerIXP.objects.all().first()
    assert ixp_stats.local_asns_members_rate == 0.2


def test_calculate_local_asns_members_rate_returns_zero_if_no_asns_in_country():
    rate = calculate_local_asns_members_rate([12345], [])

    assert rate == 0


def test_calculate_local_asns_members_rate():
    rate = calculate_local_asns_members_rate([12345], [12345, 446, 789, 5050, 54321])

    assert rate == 0.2


def test_calculate_local_asns_members_rate_ignores_members_not_in_country_list():
    rate = calculate_local_asns_members_rate([12345, 789], [12345, 446, 5050, 54321])

    assert rate == 0.25


def create_member_fixture(ixp, as_number, speed, date_left = None, member_since = None, asn_country = "CH"):
    last_active = date_left or datetime.utcnow()
    member_since = member_since or datetime(year=2024, month=4, day=1).date()
    asn = create_asn_fixture(as_number, asn_country)
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=datetime.utcnow(),
        last_active=last_active
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=member_since,
        is_rs_peer=False,
        speed=speed,
        end_date=date_left
    )
    membership.save()



import pytest
from datetime import datetime, timedelta, timezone

from ixp_tracker.importers import toggle_ixp_active_status
from ixp_tracker.models import IXP, IXPMember, IXPMembershipRecord
from tests.fixtures import create_asn_fixture, create_ixp_fixture

pytestmark = pytest.mark.django_db
processing_date = datetime.utcnow().replace(tzinfo=timezone.utc)


def test_active_ixp_with_members_remains_active():
    ixp = create_ixp_fixture(1)
    asn = create_asn_fixture(12345)
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=ixp.last_updated,
        last_active=ixp.last_active
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=ixp.created,
        is_rs_peer=True,
        speed=1000,
        end_date=None
    )
    membership.save()

    processing_date = datetime.utcnow().replace(tzinfo=timezone.utc)
    toggle_ixp_active_status(processing_date)

    ixp = IXP.objects.get(peeringdb_id=1)

    assert ixp.active_status


def test_active_ixp_with_member_marked_ended_is_marked_inactive():
    ixp = create_ixp_fixture(1)
    asn = create_asn_fixture(12345)
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=ixp.last_updated,
        last_active=ixp.last_active
    )
    member.save()
    end_date = processing_date.replace(day=1) - timedelta(days=1)
    membership = IXPMembershipRecord(
        member=member,
        start_date=ixp.created,
        is_rs_peer=True,
        speed=1000,
        end_date=end_date
    )
    membership.save()

    toggle_ixp_active_status(processing_date)

    ixp = IXP.objects.get(peeringdb_id=1)

    assert ixp.active_status is False


def test_inactive_ixp_with_no_active_members_remains_inactive():
    ixp = create_ixp_fixture(1)
    ixp.active_status = False
    ixp.save()
    asn = create_asn_fixture(12345)
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=ixp.last_updated,
        last_active=ixp.last_active
    )
    member.save()
    end_date = processing_date.replace(day=1) - timedelta(days=1)
    membership = IXPMembershipRecord(
        member=member,
        start_date=ixp.created,
        is_rs_peer=True,
        speed=1000,
        end_date=end_date
    )
    membership.save()

    toggle_ixp_active_status(processing_date)

    ixp = IXP.objects.get(peeringdb_id=1)

    assert ixp.active_status is False


def test_inactive_ixp_with_active_member_marked_active():
    ixp = create_ixp_fixture(1)
    ixp.active_status = False
    ixp.save()
    asn = create_asn_fixture(12345)
    member = IXPMember(
        ixp=ixp,
        asn=asn,
        last_updated=ixp.last_updated,
        last_active=ixp.last_active
    )
    member.save()
    membership = IXPMembershipRecord(
        member=member,
        start_date=ixp.created,
        is_rs_peer=True,
        speed=1000,
        end_date=None
    )
    membership.save()

    toggle_ixp_active_status(processing_date)

    ixp = IXP.objects.get(peeringdb_id=1)

    assert ixp.active_status



from datetime import datetime
from typing import List

import pytest

from ixp_tracker.models import StatsPerCountry
from ixp_tracker.stats import generate_stats
from tests.test_ixp_stats import create_member_fixture
from tests.test_members_import import create_ixp_fixture

pytestmark = pytest.mark.django_db


class TestLookup:

    def __init__(self, default_status: str = "assigned"):
        self.default_status = default_status

    def get_iso2_country(self, asn: int, as_at: datetime) -> str:
        pass

    def get_status(self, asn: int, as_at: datetime) -> str:
        pass

    def get_asns_for_country(self, country: str, as_at: datetime) -> List[int]:
        return [12345, 446, 789, 5050, 54321]


def test_with_no_data_generates_no_stats():
    generate_stats(TestLookup())

    stats = StatsPerCountry.objects.all()
    assert len(stats) == 249
    first_stat = stats.first()
    assert first_stat.member_count == 0


def test_generates_stats():
    ixp_one = create_ixp_fixture(123, "CH")
    create_member_fixture(ixp_one, 12345, 500)
    create_member_fixture(ixp_one, 67890, 10000)
    ixp_two = create_ixp_fixture(124, "CH")
    create_member_fixture(ixp_two, 5050, 6000)
    create_member_fixture(ixp_two, 67890, 10000)

    generate_stats(TestLookup())

    stats = StatsPerCountry.objects.filter(country_code="CH").first()
    assert stats.asn_count == 5
    assert stats.member_count == 3
    assert stats.total_capacity == 26.5
    assert stats.asns_ixp_member_rate == 0.4


def test_handles_invalid_country():
    create_ixp_fixture(123, "XK")

    generate_stats(TestLookup())

    country_stats = StatsPerCountry.objects.filter(country_code="XK").first()
    assert country_stats is None
